#---------------------------------------------------------------------------
# Cedric Adjih, Ichrak Amdouni - Hipercom Project-Team - INRIA - 2011
#---------------------------------------------------------------------------

import math
import cairo

nodeSize, nodeBorderSize, nodePointSize = (0.25, 0.25+0.02, 0.03)
maxXCoord = 14
maxYCoord = 10
minYCoord = -3

xSize = maxXCoord*2+1
ySize = maxYCoord-minYCoord+1

#--------------------------------------------------

v1 = (3,1)
v2 = (-1,2)

w0 = (8,9)

def mul(a,u): 
    return (a*u[0], a*u[1])

def plus(u,v): 
    return (u[0]+v[0], u[1]+v[1])

def det(u,v):
    return u[0]*v[1] - u[1]*v[0]

dd = det(v1,v2)

g1 = 2
g2 = 3

u1 = mul(g1,v1)
u2 = mul(g2,v2)

d = det(u1,u2)

#--------------------------------------------------

def gcd(a,b):
    a,b = abs(a),abs(b)
    if a==0: return b
    if b==0: return a
    while b > 0: a,b = b, a%b
    return a

def modulo(a,b):
    assert b>0
    assert 0<=a%b<b
    return a%b

class ColorComputationNCC2:

    def __init__(self, u1, u2):
        if det(u1,u2) < 0:
            u2 = (-u2[0],-u2[1])
            self.sign = -1
            raise FatalError("Untested")
        else: self.sign = 1
        self.u1 = u1
        self.u2 = u2
        self.g1 = gcd(*u1)
        self.g2 = gcd(*u2)
        self.v1 = (self.u1[0]//self.g1, self.u1[1]//self.g1)
        self.v2 = (self.u2[0]//self.g2, self.u2[1]//self.g2)
        self.d = det(self.u1,self.u2)
        self.dp = det(self.v1,self.v2)
        assert self.d > 0 and self.dp > 0
        print self.v1, self.v2, self.d, self.dp

    def _getCoord(self, w):
        if self.sign < 0:
            w = (w[0],-w[1])
        ap = det(w, self.v2) // self.dp
        bp = det(self.v1, w) // self.dp
        xc = modulo(ap, self.g1)
        yc = modulo(bp, self.g2)
        xx = modulo(det(w, self.v2), self.dp)
        yy = modulo(det(self.v1,w), self.dp)
        return (ap,bp), (xc,yc), (xx, yy)

    def getColor(self, w):
        (xb,yb), (xc,yc), (xx,yy) = self._getCoord(w)
        #return xc + self.g1 * yc + self.g1*self.g2 * xx
        #return (xc,yc,xx)
        return 1+ xx + self.dp*( xc + self.g1*yc  ) 



class ColorComputationNCC1:

    def __init__(self, u1, u2):
        if det(u1,u2) < 0:
            u2 = (-u2[0],-u2[1])
            self.sign = -1
            raise FatalError("Untested")
        else: self.sign = 1
        self.u1 = u1
        self.u2 = u2
        self.d = det(self.u1,self.u2)
        assert self.d > 0

        self._computeColorList()

    def _computeColorList(self):
        xList = [self.u1[0],self.u2[0]]
        xmin=min(xList)
        xmax=max(xList)
        yList = [self.u1[1],self.u2[1]]
        ymin=min(xList)
        ymax=max(xList)

        valueList = set()
        for x in range(xmin,xmax+1):
            for y in range(ymin,ymax+1):
                (c1,c2) = (modulo(det((x,y),u2), det(u1,u2)),
                           modulo(det(u1,(x,y)), det(u1,u2)))
                valueList.add((c1,c2))
        valueList = sorted(valueList)
        assert len(valueList) == self.d
        self.colorList = valueList
        #print valueList

    def getColor(self, w):
        if self.sign < 0:
            w = (w[0],-w[1])
        c1 = modulo(det(w,u2), self.d)
        c2 = modulo(det(u1,w), self.d)
        return 1+self.colorList.index((c1,c2))

#--------------------------------------------------

scale = 10
pdfImage = cairo.PDFSurface("ncc2-method.pdf", scale*xSize, scale*ySize)
cr = cairo.Context(pdfImage)

lineWidth = 0.02

#cr.set_line_width(lineWidth)

cr.scale(scale,scale)
cr.translate(+maxXCoord+0.5,+maxYCoord+0.5)
cr.transform(cairo.Matrix(1, 0, 0, -1, 0, 0))

cr.rectangle(-maxXCoord,minYCoord,2*maxXCoord,maxYCoord-minYCoord)
cr.clip()


pointSet = set([(x,y)
                for x in range(-maxXCoord,maxXCoord+1)
                for y in range(minYCoord,maxYCoord+1)])

largePointSet = set([(x,y)
                     for x in range(-2*maxXCoord,2*maxXCoord+1)
                     for y in range(-2*maxYCoord,2*maxYCoord+1)])

# ----- The colored parts
for xc in range(g1):
    for yc in range(g2):
        xx1,yy1 = plus(plus(mul(xc,v1), mul(yc,v2)),u1)
        cr.move_to(xx1,yy1)
        cr.line_to(xx1+v1[0], yy1+v1[1])
        cr.line_to(xx1+v1[0]+v2[0], yy1+v1[1]+v2[1])
        cr.line_to(xx1+v2[0], yy1+v2[1])
        cr.close_path()
        c = 0.9 - ((xc+yc*g1)/ float(g1*g2))*0.5
        cr.set_source_rgb(c,c,c)
        cr.fill()

# ----- X-Axis and Y-Axis
cr.set_source_rgb(0,0,0)
cr.set_line_width(3*lineWidth)
cr.move_to(0,0)
cr.line_to(maxXCoord,0)
cr.move_to(0,0)
cr.line_to(0,maxYCoord)
cr.move_to(0,0)
cr.line_to(-maxXCoord,0)
cr.move_to(0,0)
cr.line_to(0,minYCoord)
cr.stroke()


# ----- Lattice L(u1,u2) and L(v1,v2)
for w in largePointSet:
    if det(w,u1) % d == 0 and det(w,u2) % d == 0:
        x,y = w
        dx1,dy1 = u1
        dx2,dy2 = u2
        #cr.save()
        cr.set_line_width(6*lineWidth)
        cr.move_to(x,y)
        cr.line_to(x+dx1,y+dy1)
        cr.move_to(x,y)
        cr.line_to(x+dx2,y+dy2)
        cr.stroke()
        #cr.restore()
    elif det(w,v1) % dd == 0 and det(w,v2) % dd == 0:
        x,y = w
        dx1,dy1 = v1
        dx2,dy2 = v2
        #cr.save()
        cr.set_line_width(lineWidth)
        cr.move_to(x,y)
        cr.line_to(x+dx1,y+dy1)
        cr.move_to(x,y)
        cr.line_to(x+dx2,y+dy2)
        cr.stroke()
        #cr.restore()

# ----- Colored points

colorNCC2 = ColorComputationNCC2(u1,u2)

cSet = set()

toneList = [ (1,0,0), (0,1,1), 
             (0,1,0), (1,0,1),
             (0,0,1), (1,1,0), (1,1,1) ]

for w in pointSet:
    #cr.rectangle(x-nodeSize/2.0,y-nodeSize/2.0,nodeSize,nodeSize)
    #cr.stroke()
    (x,y) = w


    a = det(w,u2)
    b = -det(w,u1)
    if -d <= a < 0 and 0 <= b < d:
        cr.arc(x,y,nodeSize, 0, 2*math.pi)
        c = colorNCC2.getColor(w)
        cSet.add(c)

        # calc 1
        nc = dd//3
        ccc = (c-1) % dd
        baseCol = [(ccc % nc), (ccc//nc)%nc, (ccc//nc)//nc]
        baseCol = [ (x+1) / float(nc+2) for x in baseCol]
        cc = 1.0 - 0.6 * (((c-1) // dd) /float (g1*g2))
        col = tuple([cc * x for x in baseCol])

        # calc 2
        col = tuple(3*[c /float (d+2)])

        # calc 4
        ccc = (c-1) // dd
        cc = 1.0 - 0.9 * (ccc /float (dd))
        tone = (c-1) % dd
        tone = toneList[tone]
        col = tuple([x*cc for x in tone])

        # calc 3
        ccc = (c-1) % dd
        cc = 1.0 - 0.9 * (ccc /float (dd))
        tone = (c-1) // dd
        tone = toneList[tone]
        col = tuple([x*cc for x in tone])


        cr.set_source_rgb(*col)
        cr.fill()
    else: 
        cr.set_source_rgb(0,0,0)
        if (x,y) != w0:
            cr.arc(x,y,nodeSize/3.0, 0, 2*math.pi)
            cr.stroke()
        else: 
            cr.arc(x,y,nodeSize, 0, 2*math.pi)
            cr.fill()
 
assert len(cSet) == abs(det(u1,u2))


# ---- Arrow

def drawArrowHead(xy, u, angle, length):
    cr.set_source_rgb(0,0,0)
    x,y = xy
    a0 = math.atan2(u[1],u[0]) + math.pi
    for a1 in [-angle,+angle]:
        xx = length*math.cos(a0+a1) + x
        yy = length*math.sin(a0+a1) + y
        cr.move_to(x,y)
        cr.line_to(xx,yy)
        cr.stroke()
        #print x,y, xx,yy
    
cr.set_line_width(5*lineWidth)
drawArrowHead(u1, u1, 0.5, 0.7)
drawArrowHead(u2, u2, 0.5, 0.7)

cr.set_line_width(3*lineWidth)
drawArrowHead(plus(u1,v1), v1, 0.5, 0.7)
drawArrowHead(plus(u1,v2), v2, 0.5, 0.7)

# ----- Text u_1, u_2

cr.transform(cairo.Matrix(1, 0, 0, -1, 0, 0)) # hack (avoid mirror text!)
#cr.translate(0,-maxYCoord-0.5)
cr.select_font_face("Sans")


def writeU(x,y,l,c, scale=1):
    cr.set_font_size(1.5*scale)
    cr.move_to(x,y)
    cr.show_text(l)
    cr.move_to(x+0.8,y+0.3)
    cr.set_font_size(1*scale)
    cr.show_text(c)

writeU(u1[0]//2-0.7,-0.6-u1[1]//2, "u","1")
writeU(u2[0]//2+0.6,-u2[1]//2, "u","2")


writeU(u1[0]+v1[0]//2-0,0.5-v1[1]//2-u1[1], "v","1")
writeU(u1[0]+v2[0]//2+0.6,-v2[1]//2-u1[1], "v","2")


cr.set_font_size(1.5)
cr.move_to(w0[0]+0.3,-w0[1]+1)
cr.show_text("w")

cr.move_to(-1,+1.5)
cr.show_text("O")

cr.show_page()
pdfImage.finish()

#--------------------------------------------------

colorNCC1 = ColorComputationNCC1(u1,u2)

c1w,c2w = (modulo(det(w0,u2), det(u1,u2)), modulo(det(u1,w0), det(u1,u2)))
print "u1=%s, u2=%s, d=%s, w=%s" % (u1,u2,d,w0)
print "VCM-NCC1: c1=%s c2=%s c#=%d" % (c1w, c2w, colorNCC1.getColor(w0))

xx1,xx2 = det(w0, u2), det(u1,u2)
print "xu = %s/%s  [%s %s]" % (xx1/gcd(xx1,xx2),xx2/gcd(xx1,xx2), xx1,xx2),
yy1,yy2 = det(u1, w0), det(u1,u2)
print "yu = %s/%s" % (yy1/gcd(yy1,yy2),yy2/gcd(yy1,yy2))

print "v1=%s, v2=%s" % (v1,v2)
print "VCM-NCC2: (xb,yb), (xc,yc), (xx,yy)=", colorNCC2._getCoord(w0)
print "VCM-NCC2: color=%s" % colorNCC2.getColor(w0)
#print "VCM-NCC2: %s" % 

xx1,xx2 = det(w0, v2), det(v1,v2)
print "xu' = %s/%s" % (xx1/gcd(xx1,xx2),xx2/gcd(xx1,xx2)),
yy1,yy2 = det(v1, w0), det(v1,v2)
print "yu' = %s/%s" % (yy1/gcd(yy1,yy2),yy2/gcd(yy1,yy2))



#---------------------------------------------------------------------------
